import torch
import torch.nn as nn
import torch_geometric.nn as pyg_nn
from torch_geometric.data import HeteroData
from tgb.linkproppred.dataset import LinkPropPredDataset
from tgb.linkproppred.evaluate import Evaluator

class TGB2TemporalGNN(nn.Module):
    def __init__(self, num_nodes, edge_dim, hidden_dim, output_dim, num_layers):
        super(TGB2TemporalGNN, self).__init__()
        self.node_emb = nn.Embedding(num_nodes, hidden_dim)
        self.edge_encoder = nn.Linear(edge_dim, hidden_dim)
        
        self.convs = nn.ModuleList()
        self.convs.append(pyg_nn.GATConv(hidden_dim, hidden_dim, edge_dim=hidden_dim))
        for _ in range(num_layers - 2):
            self.convs.append(pyg_nn.GATConv(hidden_dim, hidden_dim, edge_dim=hidden_dim))
        self.convs.append(pyg_nn.GATConv(hidden_dim, output_dim, edge_dim=hidden_dim))
        
        self.time_encoder = nn.Linear(1, hidden_dim)
        
    def forward(self, data: HeteroData):
        x = self.node_emb(data.src)
        edge_attr = self.edge_encoder(data.edge_attr)
        edge_index = data.edge_index
        
        # 时间编码
        time_emb = self.time_encoder(data.t.unsqueeze(-1).float())
        x = x + time_emb
        
        # 图卷积
        for conv in self.convs[:-1]:
            x = conv(x, edge_index, edge_attr)
            x = torch.relu(x)
        x = self.convs[-1](x, edge_index, edge_attr)
        
        return x
    
    def link_pred_loss(self, z, pos_edge_index, neg_edge_index):
        pos_score = (z[pos_edge_index[0]] * z[pos_edge_index[1]]).sum(dim=-1)
        neg_score = (z[neg_edge_index[0]] * z[neg_edge_index[1]]).sum(dim=-1)
        return torch.nn.functional.binary_cross_entropy_with_logits(
            torch.cat([pos_score, neg_score]),
            torch.cat([torch.ones_like(pos_score), torch.zeros_like(neg_score)])
        )

# 加载TGB2.0数据集
dataset = LinkPropPredDataset(name="tgbl-wiki", root="data/")
data = dataset.get_TemporalData()
evaluator = Evaluator(name="tgbl-wiki")

# 模型初始化
model = TGB2TemporalGNN(
    num_nodes=data.num_nodes,
    edge_dim=data.edge_attr.size(1),
    hidden_dim=64,
    output_dim=32,
    num_layers=3
)

optimizer = torch.optim.Adam(model.parameters(), lr=0.01)

# 训练循环
for epoch in range(100):
    model.train()
    optimizer.zero_grad()
    
    z = model(data)
    loss = model.link_pred_loss(z, data.edge_index, data.negative_edge_index)
    
    loss.backward()
    optimizer.step()
    
    print(f"Epoch {epoch}, Loss: {loss.item()}")
    
    # 评估
    if epoch % 10 == 0:
        model.eval()
        with torch.no_grad():
            z = model(data)
            pos_pred = (z[data.edge_index[0]] * z[data.edge_index[1]]).sum(dim=-1)
            neg_pred = (z[data.negative_edge_index[0]] * z[data.negative_edge_index[1]]).sum(dim=-1)
            
            input_dict = {
                "y_pred_pos": pos_pred,
                "y_pred_neg": neg_pred,
            }
            metrics = evaluator.eval(input_dict)
            print(f"Epoch {epoch}, Metrics: {metrics}")
