import torch
from torch.optim import Adam
from tgb.linkproppred.dataset import LinkPropPredDataset
from tgb.linkproppred.evaluate import Evaluator
from TGB2TemporalGNN import TGB2TemporalGNN

# 设置设备
device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

# 加载数据集
dataset_name = "tgbl-wiki"
dataset = LinkPropPredDataset(name=dataset_name, root="data/")
data = dataset.get_TemporalData()
data = data.to(device)

# 初始化评估器
evaluator = Evaluator(name=dataset_name)

# 模型参数
num_nodes = data.num_nodes
edge_dim = data.edge_attr.size(1)
hidden_dim = 64
output_dim = 32
num_layers = 3

# 初始化模型
model = TGB2TemporalGNN(
    num_nodes=num_nodes,
    edge_dim=edge_dim,
    hidden_dim=hidden_dim,
    output_dim=output_dim,
    num_layers=num_layers
).to(device)

# 优化器
optimizer = Adam(model.parameters(), lr=0.01)

# 训练函数
def train():
    model.train()
    optimizer.zero_grad()
    z = model(data)
    loss = model.link_pred_loss(z, data.edge_index, data.negative_edge_index)
    loss.backward()
    optimizer.step()
    return loss.item()

# 评估函数
@torch.no_grad()
def evaluate():
    model.eval()
    z = model(data)
    pos_pred = (z[data.edge_index[0]] * z[data.edge_index[1]]).sum(dim=-1)
    neg_pred = (z[data.negative_edge_index[0]] * z[data.negative_edge_index[1]]).sum(dim=-1)
    
    input_dict = {
        "y_pred_pos": pos_pred.cpu(),
        "y_pred_neg": neg_pred.cpu(),
    }
    metrics = evaluator.eval(input_dict)
    return metrics

# 训练循环
num_epochs = 100
for epoch in range(num_epochs):
    loss = train()
    print(f"Epoch {epoch:03d}, Loss: {loss:.4f}")
    
    if epoch % 10 == 0:
        metrics = evaluate()
        print(f"Epoch {epoch:03d}, Metrics: {metrics}")

# 最终评估
final_metrics = evaluate()
print(f"Final Metrics: {final_metrics}")

# 保存模型
torch.save(model.state_dict(), f"tgb2_model_{dataset_name}.pth")
print("Model saved.")
